// Author : Nicolas JEAN
// Date   : October 2001
// filename : define.c
// Copyright : Nicolas JEAN 2001 - 2004
// available on http://youpee.salemioche.com

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "common.h"
#include "define.h"

char *html_code[][2]= { { "" , "&eacute;" },
                        { "" , "&egrave;" },
                        { "" , "&ecirc;"  },
                        { "" , "&agrave;" },
                        { "" , "&ccedil;" },
                        { 0x00, 0x00     } } ;

def *builtin_defines = NULL;
int bdef_nb = 0;
int bdef_lg = 0;

def *defines = NULL;
int def_nb = 0;
int def_lg = 0;

void add_to_define_array(char *s1, char *s2, def **defines, int *nb, int *lg) {
  if ( ! s1 ) return;
  if ( *nb >= *lg ) 
    *defines = (def *) realloc(*defines, (*lg += 50) *sizeof(def));

  (*defines)[*nb].from = s1;
  (*defines)[*nb].to = s2;

  (*defines)[++(*nb)].from = NULL;
}

void del_to_define_array(char *pattern, def **defines, int *nb, int *lg)
{
int i=0;
  
  if ( !(*defines) || !pattern) return;

  while ( (*defines)[i].from && strcmp(pattern, (*defines)[i].from) ) i++;

  for ( ; i < *nb ; i++ ) {
    (*defines)[i].from = (*defines)[i+1].from;
    (*defines)[i].to   = (*defines)[i+1].to;
  }

  (*nb)--; (*lg)--;
  (*defines)[i].from = NULL;
}

void del_all_define_array(def **defines, int *nb, int *lg) {
  *nb = *lg = 0;
  (*defines)[0].from = NULL;
}

void add_builtin_define(char *s1, char *s2) {
  add_to_define_array( s1, s2, &builtin_defines, &bdef_nb, &bdef_lg);
}

void replace_builtin_define(char *s1, char *s2) {
  replace_or_add_in_define_array( s1, s2, &builtin_defines, &bdef_nb, &bdef_lg);

}

void add_bindefine(FILE *in, FILE *out, char *line) {
char s[255];

  sprintf(s, BINDEF_STR, get_field(line, 2));
  replace_or_add_in_define_array( strdup(s), get_field(line, 3), &builtin_defines, &bdef_nb, &bdef_lg);

}

void add_tab(FILE *in, FILE *out, char *line) {
char fulltab[255], oneline[255], tabsizetag[255], tabsize[255];
char *tab_name, *str, *p, *field;
char *tab[255], buf[MAX_LINE_LENGTH];
int i=0, j;
unsigned int l = 0;

  sprintf(fulltab, BINDEF_STR, tab_name = get_field(line, 2));
  while ( getstrline(buf,MAX_LINE_LENGTH,in) != NULL ) {
    tab[i] = strdup(buf);
    if ( strncmp_sol(tab[i],DIESE_ENDTAB)) {
      l += strlen(tab[i]);
      sprintf(oneline,TABDEF_STR, tab_name, i+1);
      add_to_define_array( strdup(oneline), removeRN(tab[i]), &builtin_defines, &bdef_nb, &bdef_lg);
      j=1;
      while ( ( field = get_field(tab[i],j) ) ) {
        sprintf(oneline,TABLINEDEF_STR, tab_name, i+1, j);
        add_to_define_array( strdup(oneline), field, &builtin_defines, &bdef_nb, &bdef_lg);
        j++;
      }
      sprintf(tabsizetag,TABLINESIZE_STR, tab_name, i+1);
      sprintf(tabsize,"%d", j-1);
      add_to_define_array( strdup(tabsizetag), strdup(tabsize), &builtin_defines, &bdef_nb, &bdef_lg);
    } else 
      break;
    i++;
  }

  sprintf(tabsizetag,TABSIZE_STR, tab_name);
  sprintf(tabsize,"%d", i);
  add_to_define_array( strdup(tabsizetag), strdup(tabsize), &builtin_defines, &bdef_nb, &bdef_lg);
  p = str = (char *) malloc(sizeof(char)*(l+1));
  for ( j=0; j< i; p+=strlen(tab[j++]) ) strcpy(p,tab[j]);
  add_to_define_array( strdup(fulltab), str, &builtin_defines, &bdef_nb, &bdef_lg);

}

void add_list(FILE *in, FILE *out, char *line) {
char fulllist[255], oneline[255], listsizetag[255], listsize[255];
char *list_name, *str, *p;
char *tab[255], buf[MAX_LINE_LENGTH];
int i=0, j;
unsigned int l = 0;

  sprintf(fulllist, BINDEF_STR, list_name = get_field(line, 2));
  while ( getstrline(buf,MAX_LINE_LENGTH,in) != NULL ) {
    tab[i] = strdup(buf);
    if ( strncmp_sol(tab[i],DIESE_ENDLIST)) {
      l += strlen(tab[i]);
      sprintf(oneline,LISTDEF_STR, list_name, i+1);
      add_to_define_array( strdup(oneline), removeRN(tab[i]), &builtin_defines, &bdef_nb, &bdef_lg);
    } else 
      break;
    i++;
  }

  sprintf(listsizetag,LISTSIZE_STR, list_name);
  sprintf(listsize,"%d", i);
  add_to_define_array( strdup(listsizetag), strdup(listsize), &builtin_defines, &bdef_nb, &bdef_lg);
  p = str = (char *) malloc(sizeof(char)*(l+1));
  for ( j=0; j< i; p+=strlen(tab[j++]) ) strcpy(p,tab[j]);
  add_to_define_array( strdup(fulllist), str, &builtin_defines, &bdef_nb, &bdef_lg);

}

void add_var(FILE *in, FILE *out, char *line) {
char s[255], *str, *p;
char *tab[255], buf[MAX_LINE_LENGTH];
int i=0, j;
unsigned int l = 0;

  sprintf(s, BINDEF_STR, get_field(line, 2));
  while ( getstrline(buf,MAX_LINE_LENGTH,in) != NULL) {
    tab[i] = strdup(buf);
    l += strlen(tab[i]);
    if ( ! strncmp_sol(tab[i],DIESE_ENDVAR) ) break;
    i++;
  }
  p = str = (char *) malloc(sizeof(char)*(l+1));
  for ( j=0; j< i; p+=strlen(tab[j++]) ) strcpy(p,tab[j]);
  add_to_define_array( strdup(s), str, &builtin_defines, &bdef_nb, &bdef_lg);
  
}


void add_stddefine(char *s1, char *s2) {
  add_to_define_array(s1, s2, &defines, &def_nb, &def_lg);
}

void add_define(FILE *in, FILE *out, char *line) {
  
  add_to_define_array( get_field(line, 2), get_field(line, 3),
                       &defines, &def_nb, &def_lg);

}

void del_stddefine(char *pattern)
{
  del_to_define_array(pattern, &defines, &def_nb, &def_lg);
}

void del_define(FILE *in, FILE *out, char *line)
{
  del_stddefine(get_field(line, 2));
}

int is_def(char *from, def *defines)
{
int i=0;
  
  if ( !defines ) return FALSE;
  while ( defines[i].from && strcmp(from, defines[i].from) ) i++;
  if ( defines[i].from ) return TRUE;
  return FALSE;
}

int is_bindefined(char *from)
{
  return is_def(from, builtin_defines);
}
int is_defined(char *from)
{
  return is_def(from, defines);
}

void replace_in_define_array(char *from, char *to, def **defines)
{
int i=0;
  
  if ( !defines ) return;
  while ( (*defines)[i].from && strcmp(from, (*defines)[i].from) ) i++;
  if ( (*defines)[i].from )
   (*defines)[i].to = to;
  else {  // we add                    !!!!!!! memory problem is possible here if allocation has
          //                           !!!!!!! not been done yet. ( will be fixed in 4.0 ) 
   (*defines)[i].from = from;
   (*defines)[i].to = to;
  }
}

void replace_or_add_in_define_array(char *from, char *to, def **defines, int *nb, int *lg)
{
int i=0;
  if ( !defines ) return;
  while ( (*defines)[i].from && strcmp(from, (*defines)[i].from) ) i++;
  if ( (*defines)[i].from )
   (*defines)[i].to = to;
  else   
    add_to_define_array(from, to, defines, nb, lg);
}

int do_replace ( char *line , def *changes) {
def* d = changes;
int did_one, made_change=0;
  if ( ! d ) return 0;

  do { 
   did_one = 0;
   while ( d->from ) {

     if ( replace_string(line, d->from, d->to) )
      did_one++;
     d++;
   }
   d = changes;
   if (did_one) made_change++;
  } while (did_one);

  return made_change;
}

void list_replace ( FILE *in, FILE *out, char *line) {
def* d = builtin_defines;
  if ( ! d ) return;
  while ( d->from ) {
    fprintf(stderr, ">>>>>>>>> %s -> %s\n", d->from, d->to);
    d++;
  }

  return;
}


int do_defines ( char *line ) {
  return do_replace(line,defines);
}

int do_builtin_defines( char *line ) {
  return do_replace(line, builtin_defines);
}

void enable_html_encode(FILE *in, FILE *out, char *line) {
int i=0;

  while ( html_code[i][0] ) {
     add_builtin_define(html_code[i][0], html_code[i][1]);
     i++;
  }
}

void disable_html_encode(FILE *in, FILE *out, char *line) {
int i=0;

  while ( html_code[i][0] ) {
    del_to_define_array(html_code[i][0], &builtin_defines, &bdef_nb, &bdef_lg);
    i++;
  }

}
